/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 *   The contents of this file are subject to the Mozilla Public License Version
 *   1.1 (the "License"); you may not use this file except in compliance with
 *   the License. You may obtain a copy of the License at
 *   http://www.mozilla.org/MPL/
 *
 *   Software distributed under the License is distributed on an "AS IS" basis,
 *   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 *   for the specific language governing rights and limitations under the
 *   License.
 *
 *   Copyright 2009-2011, Red Hat Inc.
 *   Based on mozilla.org's scriptable plugin example
 *
 *   The Original Code is mozilla.org code.
 *
 *   The Initial Developer of the Original Code is
 *   Netscape Communications Corporation.
 *   Portions created by the Initial Developer are Copyright (C) 1998
 *   the Initial Developer. All Rights Reserved.
 *
 *   Contributor(s):
 *   Uri Lublin
 *   Martin Stransky
 *   Peter Hatina
 *
 *   Alternatively, the contents of this file may be used under the terms of
 *   either the GNU General Public License Version 2 or later (the "GPL"), or
 *   the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 *   in which case the provisions of the GPL or the LGPL are applicable instead
 *   of those above. If you wish to allow use of your version of this file only
 *   under the terms of either the GPL or the LGPL, and not to allow others to
 *   use your version of this file under the terms of the MPL, indicate your
 *   decision by deleting the provisions above and replace them with the notice
 *   and other provisions required by the GPL or the LGPL. If you do not delete
 *   the provisions above, a recipient may use your version of this file under
 *   the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <sys/types.h>
#include <sys/wait.h>

#include <stdlib.h>
#include <dlfcn.h>
#include <errno.h>
#include <unistd.h>
#include <string>
#include <sstream>
#include <signal.h>

#include "nsCOMPtr.h"

// for plugins
#ifndef XP_UNIX
#include <nsString.h>
#include <nsIDOMNavigator.h>
#include <nsIDOMPluginArray.h>
#include <nsIDOMPlugin.h>
#include <nsIDOMMimeType.h>
#include <nsIObserverService.h>
#include <plugin/nsIPluginHost.h>

static NS_DEFINE_CID(kPluginManagerCID, NS_PLUGINMANAGER_CID);
#endif

#include <nsIServiceManager.h>
#include <nsISupportsUtils.h> // some usefule macros are defined here


#include <log4cpp/PropertyConfigurator.hh>
#include <log4cpp/BasicConfigurator.hh>

#include <fstream>
#include <set>
#include "config.h"
#include "controller.h"
#include "debug.h"
#include "plugin.h"
#include "nsScriptablePeer.h"

DECLARE_NPOBJECT_CLASS_WITH_BASE(ScriptablePluginObject,
                                 AllocateScriptablePluginObject);

namespace {
    const std::string ver(PACKAGE_VERSION);
    const std::string PLUGIN_NAME = "Spice Firefox Plugin " + ver;
    const std::string MIME_TYPES_DESCRIPTION = "application/x-spice:qsc:" + PLUGIN_NAME;
    const std::string PLUGIN_DESCRIPTION = PLUGIN_NAME + " Spice Client wrapper for firefox";
    const std::string LOGGER_CONFIG = "/etc/spice/logger.ini";

    // helper function for string copy
    char *stringCopy(const std::string &src)
    {
        char *dest = static_cast<char *>(NPN_MemAlloc(src.length() + 1));
        if (dest)
            strcpy(dest, src.c_str());

        return dest;
    }
    
    // helper function for tcp/udp range conversion and validation
    static int portToInt(const std::string &port)
    {
        errno = 0;
        char *end;
        const long int min = 0;
        const long int max = 65535;
        long int conv = strtol(port.c_str(), &end, 10);
        return (errno || *end != '\0' || end == port.c_str() || conv < min || conv > max)
            ? -1 : static_cast<int>(conv);
    }
}

#ifdef NPAPI_USE_CONSTCHARS
const char *NPP_GetMIMEDescription(void)
{
    return const_cast<char *>(MIME_TYPES_DESCRIPTION.c_str());
}
#else
char *NPP_GetMIMEDescription(void)
{
    return strdup(MIME_TYPES_DESCRIPTION.c_str());
}
#endif

//////////////////////////////////////
//
// general initialization and shutdown
//
NPError NS_PluginInitialize()
{
    return NPERR_NO_ERROR;
}

void NS_PluginShutdown()
{
}

// get values per plugin
NPError NS_PluginGetValue(NPPVariable aVariable, void *aValue)
{
    switch (aVariable)
    {
    case NPPVpluginNameString:
        *(static_cast<char **>(aValue)) = const_cast<char *>(PLUGIN_NAME.c_str());
        break;

    case NPPVpluginDescriptionString:
        *(static_cast<char **>(aValue)) = const_cast<char *>(PLUGIN_DESCRIPTION.c_str());
        break;

    default:
        return NPERR_INVALID_PARAM;
    }

    return NPERR_NO_ERROR;
}


/////////////////////////////////////////////////////////////
//
// construction and destruction of our plugin instance object
//
nsPluginInstanceBase *NS_NewPluginInstance(nsPluginCreateData *aCreateDataStruct)
{
    if(!aCreateDataStruct)
        return NULL;

    nsPluginInstance *plugin = new nsPluginInstance(aCreateDataStruct->instance);

    // now is the time to tell Mozilla that we are windowless
    NPN_SetValue(aCreateDataStruct->instance, NPPVpluginWindowBool, NULL);

    return plugin;
}

void NS_DestroyPluginInstance(nsPluginInstanceBase *aPlugin)
{
    if (aPlugin)
        delete static_cast<nsPluginInstance *>(aPlugin);
}

////////////////////////////////////////
//
// nsPluginInstance class implementation
//

nsPluginInstance::nsPluginInstance(NPP aInstance):
    nsPluginInstanceBase(),
    m_pid_controller(-1),
    m_pid_usbrdrctrl(-1),
    m_connected_status(-2),
    m_instance(aInstance),
    m_initialized(PR_TRUE),
    m_window(NULL),
    m_fullscreen(PR_FALSE),
    m_smartcard(PR_FALSE),
    m_admin_console(PR_FALSE),
    m_no_taskmgr_execution(PR_FALSE),
    m_send_ctrlaltdel(PR_TRUE),
    m_usb_port(0),
    m_usb_auto_share(PR_TRUE),
    m_scriptable_peer(NULL)
{
    // create .spicec directory in $HOME
    m_home_dir = getenv("HOME");
    m_home_dir += "/.spicec";
    mkdir(m_home_dir.c_str(), S_IRWXU);

    // create temporary directory in /tmp
    char tmp_dir[] = "/tmp/spicec-XXXXXX";
    m_tmp_dir = mkdtemp(tmp_dir);

    // configure log4cpp
    std::ifstream log_init(LOGGER_CONFIG.c_str());
    if (log_init.good())
    {
        log_init.close();
        log4cpp::PropertyConfigurator::configure(LOGGER_CONFIG.c_str());
    }
    else
    {
        log4cpp::BasicConfigurator::configure();
    }
}

nsPluginInstance::~nsPluginInstance()
{
    // m_scriptable_peer may be also held by the browser
    // so releasing it here does not guarantee that it is over
    // we should take precaution in case it will be called later
    // and zero its m_plugin member
    if (m_scriptable_peer)
        NPN_ReleaseObject(m_scriptable_peer);

    // delete the temporary directory used for a client socket
    rmdir(m_tmp_dir.c_str());
}

NPBool nsPluginInstance::init(NPWindow *aWindow)
{
    m_initialized = PR_TRUE;

    m_host_ip.clear();
    m_port.clear();
    m_password.clear();
    m_secure_port.clear();
    m_cipher_suite.clear();
    m_ssl_channels.clear();
    m_trust_store.clear();
    m_host_subject.clear();
    m_title.clear();
    m_dynamic_menu.clear();
    m_number_of_monitors.clear();
    m_guest_host_name.clear();
    m_hot_keys.clear();
    m_usb_filter.clear();
    m_language.clear();
    m_trust_store_file.clear();
    m_foreign_file.clear();
    m_color_depth.clear();
    m_disable_effects.clear();
    m_proxy.clear();

    m_fullscreen = PR_FALSE;
    m_smartcard = PR_FALSE;
    m_admin_console = PR_FALSE;
    m_no_taskmgr_execution = PR_FALSE;
    m_send_ctrlaltdel = PR_TRUE;

    return m_initialized;
}

NPError nsPluginInstance::SetWindow(NPWindow *aWindow)
{
    // keep window parameters
    m_window = aWindow;
    return NPERR_NO_ERROR;
}

void nsPluginInstance::shut()
{
    m_initialized = PR_FALSE;
}

NPBool nsPluginInstance::isInitialized()
{
    return m_initialized;
}

/* attribute string hostIP; */
char *nsPluginInstance::GetHostIP() const
{
    LOG_DEBUG(m_host_ip);
    return stringCopy(m_host_ip);
}

void nsPluginInstance::SetHostIP(const char *aHostIP)
{
    m_host_ip = aHostIP;
    LOG_DEBUG(m_host_ip);
}

/* attribute string port; */
char *nsPluginInstance::GetPort() const
{
    LOG_DEBUG(m_port);
    return stringCopy(m_port);
}

void nsPluginInstance::SetPort(const char *aPort)
{
    m_port = aPort;
    LOG_DEBUG(m_port);
}

/* attribute string SecurePort; */
char *nsPluginInstance::GetSecurePort() const
{
    LOG_DEBUG(m_secure_port)
    return stringCopy(m_secure_port);
}

void nsPluginInstance::SetSecurePort(const char *aSecurePort)
{
    m_secure_port = aSecurePort;
    LOG_DEBUG(m_secure_port);
}

/* attribute string Password; */
char *nsPluginInstance::GetPassword() const
{
    LOG_DEBUG("Password requested");
    return stringCopy(m_password);
}

void nsPluginInstance::SetPassword(const char *aPassword)
{
    m_password = aPassword;
    LOG_DEBUG("Password set");
}

/* attribute string CipherSuite; */
char *nsPluginInstance::GetCipherSuite() const
{
    LOG_DEBUG(m_cipher_suite);
    return stringCopy(m_cipher_suite);
}

void nsPluginInstance::SetCipherSuite(const char *aCipherSuite)
{
    m_cipher_suite = aCipherSuite;
    LOG_DEBUG(m_cipher_suite);
}

/* attribute string SSLChannels; */
char *nsPluginInstance::GetSSLChannels() const
{
    LOG_DEBUG(m_ssl_channels);
    return stringCopy(m_ssl_channels);
}

void nsPluginInstance::SetSSLChannels(const char *aSSLChannels)
{
    m_ssl_channels = aSSLChannels;
    LOG_DEBUG("original channels: " << m_ssl_channels);

    /*
     * Backward Compatibility: Begin
     * Remove leading 's' from m_ssl_channels, e.g. "main" not "smain"
     * RHEL5 uses 'smain' and 'sinpusts
     * RHEL6 uses 'main'  and 'inputs'
     */
    const char* chan_names[] = {
        "smain", "sdisplay", "sinputs",
        "scursor", "splayback", "srecord",
        "susbredir", "ssmartcard", "stunnel"
    };
    const int nnames = sizeof(chan_names) / sizeof(chan_names[0]);

    for (int i = 0; i < nnames; i++) {
        const char *name = chan_names[i];
        size_t found = 0;
        while ((found = m_ssl_channels.find(name, found)) != std::string::npos)
            m_ssl_channels.replace(found, strlen(name), name + 1);
    }
    /* Backward Compatibility: End */

    LOG_DEBUG("modified channels: " << m_ssl_channels);
}

//* attribute string TrustStore; */
char *nsPluginInstance::GetTrustStore() const
{
    LOG_DEBUG(m_trust_store);
    return stringCopy(m_trust_store);
}

void nsPluginInstance::SetTrustStore(const char *aTrustStore)
{
    m_trust_store = aTrustStore;
    LOG_DEBUG(m_trust_store);
}

/* attribute string HostSubject; */
char *nsPluginInstance::GetHostSubject() const
{
    LOG_DEBUG(m_host_subject);
    return stringCopy(m_host_subject);
}

void nsPluginInstance::SetHostSubject(const char *aHostSubject)
{
    m_host_subject = aHostSubject;
    LOG_DEBUG(m_host_subject);
}

/* attribute boolean fullScreen; */
PRBool nsPluginInstance::GetFullScreen() const
{
    LOG_DEBUG(m_fullscreen);
    return m_fullscreen;
}

void nsPluginInstance::SetFullScreen(PRBool aFullScreen)
{
    m_fullscreen = aFullScreen;
    LOG_DEBUG(m_fullscreen);
}

/* attribute boolean Smartcard; */
PRBool nsPluginInstance::GetSmartcard() const
{
    return m_smartcard;
}

void nsPluginInstance::SetSmartcard(PRBool aSmartcard)
{
    m_smartcard = aSmartcard;
}

/* attribute string Title; */
char *nsPluginInstance::GetTitle() const
{
    LOG_DEBUG(m_title);
    return stringCopy(m_title);
}

void nsPluginInstance::SetTitle(const char *aTitle)
{
    m_title = aTitle;
    // duplicate "%", so log4cpp can properly interpret
    // it as percent symbol
    std::string title(m_title);
    size_t found = -2;
    while ((found = title.find("%", found + 2)) != std::string::npos)
        title.replace(found, 1, "%%");
    LOG_DEBUG(title);
}

/* attribute string dynamicMenu; */
char *nsPluginInstance::GetDynamicMenu() const
{
    LOG_DEBUG(m_dynamic_menu);
    return stringCopy(m_dynamic_menu);
}

void nsPluginInstance::SetDynamicMenu(const char *aDynamicMenu)
{
    m_dynamic_menu = aDynamicMenu;
    LOG_DEBUG(m_dynamic_menu);
}

/* attribute string NumberOfMonitors; */
char *nsPluginInstance::GetNumberOfMonitors() const
{
    LOG_DEBUG(m_number_of_monitors);
    return stringCopy(m_number_of_monitors);
}

void nsPluginInstance::SetNumberOfMonitors(const char *aNumberOfMonitors)
{
    m_number_of_monitors = aNumberOfMonitors;
    LOG_DEBUG(m_number_of_monitors);
}

/* attribute boolean AdminConsole; */
PRBool nsPluginInstance::GetAdminConsole() const
{
    LOG_DEBUG(m_admin_console);
    return m_admin_console;
}

void nsPluginInstance::SetAdminConsole(PRBool aAdminConsole)
{
    m_admin_console = aAdminConsole;
    LOG_DEBUG(m_admin_console);
}

/* attribute string GuestHostName; */
char *nsPluginInstance::GetGuestHostName() const
{
    LOG_DEBUG(m_guest_host_name);
    return stringCopy(m_guest_host_name);
}

void nsPluginInstance::SetGuestHostName(const char *aGuestHostName)
{
    m_guest_host_name = aGuestHostName;
    LOG_DEBUG(m_guest_host_name);
}

/* attribute string HotKey; */
char *nsPluginInstance::GetHotKeys() const
{
    LOG_DEBUG(m_hot_keys);
    return stringCopy(m_hot_keys);
}

void nsPluginInstance::SetHotKeys(const char *aHotKeys)
{
    m_hot_keys = aHotKeys;
    LOG_DEBUG(m_hot_keys);
}

/* attribute boolean NoTaskMgrExecution; */
PRBool nsPluginInstance::GetNoTaskMgrExecution() const
{
    LOG_DEBUG(m_no_taskmgr_execution);
    return m_no_taskmgr_execution;
}

void nsPluginInstance::SetNoTaskMgrExecution(PRBool aNoTaskMgrExecution)
{
    m_no_taskmgr_execution = aNoTaskMgrExecution;
    LOG_DEBUG(m_no_taskmgr_execution);
}

/* attribute boolean SendCtrlAltDelete; */
PRBool nsPluginInstance::GetSendCtrlAltDelete() const
{
    LOG_DEBUG(m_send_ctrlaltdel);
    return m_send_ctrlaltdel;
}

void nsPluginInstance::SetSendCtrlAltDelete(PRBool aSendCtrlAltDelete)
{
    m_send_ctrlaltdel = aSendCtrlAltDelete;
    LOG_DEBUG(m_send_ctrlaltdel);
}

/* attribute unsigned short UsbListenPort; */
unsigned short nsPluginInstance::GetUsbListenPort() const
{
    LOG_DEBUG(m_usb_port);
    return m_usb_port;
}

void nsPluginInstance::SetUsbListenPort(unsigned short aUsbPort)
{
    m_usb_port = aUsbPort;
    LOG_DEBUG(m_usb_port);
}

/* attribute boolean UsbAutoShare; */
PRBool nsPluginInstance::GetUsbAutoShare() const
{
    LOG_DEBUG(m_usb_auto_share);
    return m_usb_auto_share;
}

void nsPluginInstance::SetUsbAutoShare(PRBool aUsbAutoShare)
{
    m_usb_auto_share = aUsbAutoShare;
    LOG_DEBUG(m_usb_auto_share);
}

/* attribute string ColorDepth; */
char *nsPluginInstance::GetColorDepth() const
{
    return stringCopy(m_color_depth);
}

void nsPluginInstance::SetColorDepth(const char *aColorDepth)
{
    m_color_depth = aColorDepth;
}

/* attribute string DisableEffects; */
char *nsPluginInstance::GetDisableEffects() const
{
    return stringCopy(m_disable_effects);
}

void nsPluginInstance::SetDisableEffects(const char *aDisableEffects)
{
    m_disable_effects = aDisableEffects;
}

/* attribute string Proxy; */
char *nsPluginInstance::GetProxy() const
{
    return stringCopy(m_proxy);
}

void nsPluginInstance::SetProxy(const char *aProxy)
{
    m_proxy = aProxy;
}

void nsPluginInstance::WriteToPipe(const void *data, uint32_t size)
{
    m_external_controller.Write(data, size);
}

void nsPluginInstance::SendInit()
{
    ControllerInit msg = { {CONTROLLER_MAGIC, CONTROLLER_VERSION, sizeof(msg)},
                           0, CONTROLLER_FLAG_EXCLUSIVE };
    WriteToPipe(&msg, sizeof(msg));
}

void nsPluginInstance::SendMsg(uint32_t id)
{
    ControllerMsg msg = {id, sizeof(msg)};
    WriteToPipe(&msg, sizeof(msg));
}

void nsPluginInstance::SendBool(uint32_t id, PRBool value)
{
    ControllerValue msg = { {id, sizeof(msg)}, !!value };
    WriteToPipe(&msg, sizeof(msg));
}

void nsPluginInstance::SendValue(uint32_t id, uint32_t value)
{
    if (!value)
        return;

    ControllerValue msg = { {id, sizeof(msg)}, value };
    WriteToPipe(&msg, sizeof(msg));
}

void nsPluginInstance::SendBool(uint32_t id, bool value)
{
    ControllerValue msg = { {id, sizeof(msg)}, value };
    WriteToPipe(&msg, sizeof(msg));
}

void nsPluginInstance::SendStr(uint32_t id, const char *str)
{
    if (!strlen(str))
        return;

    size_t size = sizeof(ControllerData) + strlen(str) + 1;
    ControllerData *msg = static_cast<ControllerData *>(malloc(size));
    msg->base.id = id;
    msg->base.size = size;
    strcpy(reinterpret_cast<char *>(msg->data), str);
    WriteToPipe(msg, size);
    free(msg);
}

void nsPluginInstance::SendWStr(uint32_t id, const wchar_t *str)
{
    size_t len = wcslen(str);
    if (!len)
        return;

    size_t size = sizeof(ControllerData) + (len + 1) * sizeof(wchar_t);
    ControllerData *msg = static_cast<ControllerData *>(malloc(size));
    msg->base.id = id;
    msg->base.size = size;
    wcscpy(reinterpret_cast<wchar_t *>(msg->data), str);
    WriteToPipe(msg, size);
    free(msg);
}

void nsPluginInstance::Connect()
{
    const int port = portToInt(m_port);
    const int sport = portToInt(m_secure_port);
    if (port < 0)
        LOG_WARN("invalid port: '" << m_port << "'");
    if (sport < 0)
        LOG_WARN("invalid secure port: '" << m_secure_port << "'");
    if (port <= 0 && sport <= 0)
    {
        m_connected_status = 1;
        CallOnDisconnected(m_connected_status);
        return;
    }

    std::string socket_file(m_tmp_dir);
    socket_file += "/spice-xpi";
    if (setenv("SPICE_XPI_SOCKET", socket_file.c_str(), 1))
    {
        LOG_ERROR ("could not set SPICE_XPI_SOCKET env variable");
        return;
    }
    LOG_INFO("SPICE_XPI_SOCKET: " << socket_file);
    if (!m_proxy.empty()) {
        setenv("SPICE_PROXY", m_proxy.c_str(), 1);
        LOG_INFO("SPICE_PROXY: " << m_proxy);
    }

    std::string foreign_file(m_tmp_dir);
    foreign_file += "/spice-foreign";
    if (setenv("SPICE_FOREIGN_MENU_SOCKET", foreign_file.c_str(), 1))
    {
        LOG_ERROR ("could not set SPICE_FOREIGN_MENU_SOCKET env variable");
        return;
    }
    LOG_INFO("SPICE_FOREIGN_MENU_SOCKET: " << foreign_file);
    m_foreign_file = foreign_file;

    m_pid_controller = fork();
    if (m_pid_controller == 0)
    {
        setpgrp();

        char *spice_xpi_client_args[] = {
            const_cast<char*>("/usr/libexec/spice-xpi-client"),
            NULL
        };
        std::stringstream ss;
        int cnt = sizeof(spice_xpi_client_args) / sizeof(spice_xpi_client_args[0]);
        for (int i = 0; i < cnt; ++i)
            ss << spice_xpi_client_args[i] << " ";
        LOG_INFO("Launching " << ss.str());
        execv(spice_xpi_client_args[0], spice_xpi_client_args);
        LOG_ERROR("Failed to launch spice-xpi-client");

        // TODO: temporary fallback for backward compatibility
        char *spicec_args[] = {
            const_cast<char*>("/usr/bin/spicec"),
            const_cast<char*>("--controller"),
            NULL
        };
        ss.str(std::string());
        ss.clear();
        cnt = sizeof(spicec_args) / sizeof(spicec_args[0]);
        for (int i = 0; i < cnt; ++i)
            ss << spicec_args[i] << " ";
        LOG_INFO("Launching " << ss.str());
        execv(spicec_args[0], spicec_args);
        LOG_ERROR("Failed to launch spicec fallback");

        exit(1);
    }
    else
    {
        LOG_DEBUG("Controller pid: " << m_pid_controller);
        m_external_controller.SetFilename(socket_file);

        if (m_external_controller.Connect(10) != 0)
        {
            LOG_ERROR ("could not connect to spice client controller");
            return;
        }

        // create trust store filename
        FILE *fp;
        int fd = -1;
        char trust_store_template[] = "/tmp/truststore.pem-XXXXXX";
        mode_t prev_umask = umask(0177);
        fd = mkstemp(trust_store_template);
        umask(prev_umask);
        m_trust_store_file = trust_store_template;

        if (fd != -1)
        {
            fp = fdopen(fd,"w+");
            if (fp != NULL)
            {
                fputs(m_trust_store.c_str(), fp);
                fflush(fp);
                fsync(fd);
                fclose(fp);
            }
            else
            {
                LOG_ERROR ("could not open truststore temp file");
                close(fd);
                unlink(m_trust_store_file.c_str());
                m_trust_store_file.clear();
                return;
            }
        }
        else
        {
            LOG_ERROR ("could not create truststore temp file. errno=" << errno);
            return;
        }

        LOG_INFO("Initiating connection with controller");
        SendInit();
        SendStr(CONTROLLER_HOST, m_host_ip.c_str());
        if (port > 0)
            SendValue(CONTROLLER_PORT, port);
        if (sport > 0)
            SendValue(CONTROLLER_SPORT, sport);
        SendValue(CONTROLLER_FULL_SCREEN,
                   (m_fullscreen == PR_TRUE ? CONTROLLER_SET_FULL_SCREEN : 0) |
                   (m_admin_console == PR_FALSE ? CONTROLLER_AUTO_DISPLAY_RES : 0));
        SendValue(CONTROLLER_ENABLE_SMARTCARD, m_smartcard == PR_TRUE ? 1 : 0);
        SendStr(CONTROLLER_PASSWORD, m_password.c_str());
        SendStr(CONTROLLER_TLS_CIPHERS, m_cipher_suite.c_str());
        SendStr(CONTROLLER_SET_TITLE, m_title.c_str());
        SendBool(CONTROLLER_SEND_CAD, m_send_ctrlaltdel);

        SendStr(CONTROLLER_SECURE_CHANNELS, m_ssl_channels.c_str());
        SendStr(CONTROLLER_CA_FILE, m_trust_store_file.c_str());
        SendStr(CONTROLLER_HOST_SUBJECT, m_host_subject.c_str());
        SendStr(CONTROLLER_HOTKEYS, m_hot_keys.c_str());
        SendValue(CONTROLLER_COLOR_DEPTH, atoi(m_color_depth.c_str()));
        SendStr(CONTROLLER_DISABLE_EFFECTS, m_disable_effects.c_str());
        SendMsg(CONTROLLER_CONNECT);
        SendMsg(CONTROLLER_SHOW);

        // set connected status
        m_connected_status = -1;

        pthread_t controller_thread_id;
        pthread_create(&controller_thread_id, NULL, ControllerWaitHelper,
            reinterpret_cast<void *>(this));

        if (m_usb_port != 0) {
            /* old USB, disable new USB solution */
            SendBool(CONTROLLER_ENABLE_USB, PR_FALSE);
            ExecuteUsbRdrCtrl();
        } else {
            SendBool(CONTROLLER_ENABLE_USB, PR_TRUE);
            SendBool(CONTROLLER_ENABLE_USB_AUTOSHARE, m_usb_auto_share == PR_TRUE ? 1 : 0);
            SendStr(CONTROLLER_USB_FILTER, m_usb_filter.c_str());
        }
    }
}

void nsPluginInstance::Show()
{
    LOG_DEBUG("sending show message");
    SendMsg(CONTROLLER_SHOW);
}

void nsPluginInstance::Disconnect()
{
    LOG_INFO("Disconnecting");
    if (m_pid_controller > 0)
        kill(-m_pid_controller, SIGTERM);
    if (m_pid_usbrdrctrl > 0)
        kill(m_pid_usbrdrctrl, SIGTERM);
}

void nsPluginInstance::ConnectedStatus(PRInt32 *retval)
{
    *retval = m_connected_status;
}

void nsPluginInstance::SetLanguageStrings(const char *aSection, const char *aLanguage)
{
    if (aSection != NULL && aLanguage != NULL)
    {
        if (strlen(aSection) > 0 && strlen(aLanguage) > 0)
        {
            m_language[aSection] = aLanguage;
            LOG_DEBUG(m_language[aSection]);
        }
    }
}

void nsPluginInstance::SetUsbFilter(const char *aUsbFilter)
{
    if (aUsbFilter != NULL)
    {
        m_usb_filter = aUsbFilter;
        LOG_DEBUG(m_usb_filter);
    }
}

void nsPluginInstance::ExecuteUsbRdrCtrl()
{
    if (m_pid_usbrdrctrl == 0 || m_usb_port == 0)
        return;

    if ((m_pid_usbrdrctrl = fork()) == 0)
    {    
        std::stringstream ss;
        std::string log_name;
        ss << m_home_dir << "/usbrdrctrl-" << getpid() << ".log";
        ss >> log_name;
        
        int logfd = open(log_name.c_str(), O_RDWR | O_CREAT,
            S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

        if (logfd > 0)
            dup2(logfd, 1);
        
        LOG_DEBUG("Connection name: " << m_foreign_file);

        std::string port;
        ss.str(std::string());
        ss.clear();
        ss << m_usb_port;
        ss >> port;

        sleep(2);
        
        char *usbrdrctrl_args[] = {
            const_cast<char*>("/usr/bin/usbrdrctrl"),
            const_cast<char*>(m_guest_host_name.c_str()),
            const_cast<char*>(port.c_str()),
            const_cast<char*>("-c"),
            const_cast<char*>(m_foreign_file.c_str()),
            const_cast<char*>(!m_language["USB"].empty() ? "-l" : ""),
            const_cast<char*>(!m_language["USB"].empty() ? m_language["USB"].c_str() : ""),
            const_cast<char*>(!m_usb_filter.empty() ? "-f" : ""),
            const_cast<char*>(!m_usb_filter.empty() ? m_usb_filter.c_str() : ""),
            const_cast<char*>((m_usb_auto_share == PR_TRUE) ? "-a" : ""),
            NULL
        };

        ss.str(std::string());
        ss.clear();
        const int cnt = sizeof(usbrdrctrl_args) / sizeof(usbrdrctrl_args[0]);
        for (int i = 0; i < cnt; ++i)
            ss << usbrdrctrl_args[i] << " ";
        LOG_INFO("Launching " << ss.str());
        execv(usbrdrctrl_args[0], usbrdrctrl_args);
        LOG_ERROR("Failed to launch usbrdrctrl: " << errno << ": " << strerror(errno));
        exit(0);
    }
    else if (m_pid_usbrdrctrl > 0)
    {
        pthread_t thread_usbrdrctrl_id;
        pthread_create(&thread_usbrdrctrl_id, NULL, UsbRdrCtrlWaitHelper,
            reinterpret_cast<void *>(this));
    }
}

void nsPluginInstance::CallOnDisconnected(int code)
{
    NPObject *window = NULL;
    if (NPN_GetValue(m_instance, NPNVWindowNPObject, &window) != NPERR_NO_ERROR)
    {
        LOG_ERROR("could not get browser window, when trying to call OnDisconnected");
        return;
    }

    // get OnDisconnected callback
    NPIdentifier id_on_disconnected = NPN_GetStringIdentifier("OnDisconnected");
    if (!id_on_disconnected)
    {
        LOG_ERROR("could not find OnDisconnected identifier");
        return;
    }

    NPVariant var_on_disconnected;
    if (!NPN_GetProperty(m_instance, window, id_on_disconnected, &var_on_disconnected))
    {
        LOG_ERROR("could not get OnDisconnected function");
        return;
    }

    if (!NPVARIANT_IS_OBJECT(var_on_disconnected))
    {
        LOG_ERROR("OnDisconnected is not object");
        return;
    }

    NPObject *call_on_disconnected = NPVARIANT_TO_OBJECT(var_on_disconnected);

    // call OnDisconnected
    NPVariant arg;
    NPVariant void_result;
    INT32_TO_NPVARIANT(code, arg);
    NPVariant args[] = { arg };

    if (NPN_InvokeDefault(m_instance, call_on_disconnected, args, sizeof(args) / sizeof(args[0]), &void_result))
    {
        LOG_DEBUG("OnDisconnected successfuly called");
    }
    else
    {
        LOG_ERROR("could not call OnDisconnected");
    }

    // cleanup
    NPN_ReleaseObject(window);
    NPN_ReleaseVariantValue(&var_on_disconnected);
}

void *nsPluginInstance::ControllerWaitHelper(void *opaque)
{
    nsPluginInstance *fake_this = reinterpret_cast<nsPluginInstance *>(opaque);
    if (!fake_this)
        return NULL;

    int exit_code;
    waitpid(fake_this->m_pid_controller, &exit_code, 0);
    LOG_DEBUG("Controller finished, pid: " << fake_this->m_pid_controller 
        << ", exit code: " << exit_code);
    fake_this->m_connected_status = fake_this->m_external_controller.TranslateRC(exit_code);
    fake_this->CallOnDisconnected(exit_code);
    fake_this->m_external_controller.Disconnect();
    fake_this->m_pid_controller = -1;
    unlink(fake_this->m_trust_store_file.c_str());
    unlink(fake_this->m_foreign_file.c_str());
    fake_this->m_trust_store_file.clear();
    fake_this->m_foreign_file.clear();
    return NULL;
}

void *nsPluginInstance::UsbRdrCtrlWaitHelper(void *opaque)
{
    nsPluginInstance *fake_this= reinterpret_cast<nsPluginInstance *>(opaque);
    if (!fake_this)
        return NULL;

    int exit_code;
    waitpid(fake_this->m_pid_usbrdrctrl, &exit_code, 0);
    LOG_DEBUG("usbrdrctrl finished, pid: " << fake_this->m_pid_usbrdrctrl
        << ", exit code: " << exit_code);
    fake_this->m_pid_usbrdrctrl = -1;
    
    if (exit_code != 0)
        fake_this->ExecuteUsbRdrCtrl();
    
    return NULL;
}

// ==============================
// ! Scriptability related code !
// ==============================
//
// here the plugin is asked by Mozilla to tell if it is scriptable
// we should return a valid interface id and a pointer to 
// nsScriptablePeer interface which we should have implemented
// and which should be defined in the corressponding *.xpt file
// in the bin/components folder
NPError nsPluginInstance::GetValue(NPPVariable aVariable, void *aValue)
{
    // Here we indicate that the plugin is scriptable. See this page for details:
    // https://developer.mozilla.org/en/Gecko_Plugin_API_Reference/Scripting_plugins
    if (aVariable == NPPVpluginScriptableNPObject)
    {
        // addref happens in getter, so we don't addref here
        *(static_cast<NPObject **>(aValue)) = GetScriptablePeer();
    }
    return NPERR_NO_ERROR;
}

// ==============================
// ! Scriptability related code !
// ==============================
//
// this method will return the scriptable object (and create it if necessary)
NPObject *nsPluginInstance::GetScriptablePeer()
{
    if (!m_scriptable_peer)
        m_scriptable_peer = NPN_CreateObject(m_instance, GET_NPOBJECT_CLASS(ScriptablePluginObject));

    if (m_scriptable_peer)
        NPN_RetainObject(m_scriptable_peer);

    return m_scriptable_peer;
}
